"""Tests for the writing rules."""
import warnings

import pytest

import provenancetoolbox as ptb
from cirrusvolume import rules


SOURCES = ["A", "B", "C"]
MOTIVATION = "Trying to improve"
PROCESS = ptb.Process(
    "testing CirrusVolume", {"patience for humor": 3 / 10}, ptb.PythonGithubEnv(".")
)


class TestCheckWritingRules:
    """Tests whether rules.check_writing_rules enforces what we want."""

    def test_basic(self):
        """Does the function pass with the expected args."""
        rules.check_writing_rules(SOURCES, MOTIVATION, PROCESS)

    def test_nones(self):
        """Does the function raise an AssertionError with Nones."""
        with pytest.raises(AssertionError):
            rules.check_writing_rules(None, MOTIVATION, PROCESS)

        with pytest.raises(AssertionError):
            rules.check_writing_rules(SOURCES, None, PROCESS)

        with pytest.raises(AssertionError):
            rules.check_writing_rules(SOURCES, MOTIVATION, None)

    def test_types(self):
        """Does the function raise an AssertionError with incorrect types."""
        # sources not a list of strings
        with pytest.raises(AssertionError):
            rules.check_writing_rules(0, MOTIVATION, PROCESS)

        with pytest.raises(AssertionError):
            rules.check_writing_rules([0], MOTIVATION, PROCESS)

        # motivation not a string
        with pytest.raises(AssertionError):
            rules.check_writing_rules(SOURCES, 0, PROCESS)

        # process not a ptb.Process
        with pytest.raises(AssertionError):
            rules.check_writing_rules(SOURCES, MOTIVATION, 0)


class TestAddSources:
    """Tests for rules.add_sources."""

    def test_new_sources(self, readvolume):
        """Does writing sources to an empty volume work?"""
        rules.add_sources(readvolume, SOURCES)
        newsources = readvolume.provenance.sources

        assert set(newsources) == set(SOURCES)

    def test_intersection(self, readwritevolume):
        """Are duplicate sources written?"""
        sources = readwritevolume.provenance.sources

        rules.add_sources(readwritevolume, sources)

        assert len(sources) == len(readwritevolume.provenance.sources)


class TestAddMotivation:
    """Tests for rules.add_motivation."""

    def test_new_motivation(self, readvolume):
        """Does writing the motivation to an empty volume work?"""
        assert ptb.note_absent(readvolume, MOTIVATION, ptb.NoteType.MOTIVATION)

        rules.add_motivation(readvolume, MOTIVATION)

        assert not ptb.note_absent(readvolume, MOTIVATION, ptb.NoteType.MOTIVATION)

    def test_intersection(self, readwritevolume):
        """Are duplicate motivations written?"""
        assert ptb.note_absent(readwritevolume, MOTIVATION, ptb.NoteType.MOTIVATION)

        num_orig_notes = len(ptb.parsenotes(readwritevolume))
        rules.add_motivation(readwritevolume, MOTIVATION)
        num_new_notes = len(ptb.parsenotes(readwritevolume))

        assert num_new_notes == num_orig_notes + 1

        rules.add_motivation(readwritevolume, MOTIVATION)
        num_newer_notes = len(ptb.parsenotes(readwritevolume))

        assert num_newer_notes == num_new_notes


class TestAddProcess:
    """Tests for rules.add_process."""

    def test_new_process(self, readvolume):
        """Does writing the motivation to an empty volume work?"""
        assert ptb.process_absent(readvolume, PROCESS)

        rules.add_process(readvolume, PROCESS)

        assert not ptb.process_absent(readvolume, PROCESS)

    def test_intersection(self, readwritevolume):
        """Are duplicate processes written?"""
        assert ptb.process_absent(readwritevolume, PROCESS)

        num_orig_procs = len(readwritevolume.provenance.processing)
        rules.add_process(readwritevolume, PROCESS)
        num_new_procs = len(readwritevolume.provenance.processing)

        assert num_new_procs == num_orig_procs + 1

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            rules.add_process(readwritevolume, PROCESS)
        num_newer_procs = len(readwritevolume.provenance.processing)

        assert num_newer_procs == num_new_procs



"""A library of functions for enforcing and enacting the documentation rules.

THE RULES
You cannot write to a CirrusVolume unless you've passed it:
(1) A set of sources from which this was created (e.g., another CloudVolume
path or a free-form justification like "tracer annotation"). This must be
formatted as a list[str], and any sources that haven't been previously logged
will be added to the current sources field of the provenance file

(2) The motivation for creating or modifying the volume (str) has been logged and you've
included that motivation when instantiating the object. A volume can have
multiple motivation notes, and your motivation only need match one of them.

(3) You've passed a Process (a code environment & parameters) to the class.
The process will be logged unless another process with the same task
description and parameters has already been logged. These are defined by
provenancetoolbox.
"""
from __future__ import annotations

import warnings
from typing import Optional

import cloudvolume as cv
import provenancetoolbox as ptb


def check_writing_rules(
    sources: Optional[list[str]] = None,
    motivation: Optional[str] = None,
    process: Optional[ptb.Process] = None,
) -> None:
    """Checks the rules for write-access to a CloudVolume.

    Checks whether the provided fields are sufficient to allow writing
    to a volume.

    Args:
        sources: A list of descriptions for where this volume came from.
        motivation: The motivation for creating this volume.
        process: The process that describes the volume's creation.

    Raises:
        AssertionError: One of the write rules has been violated.
    """
    defined = [v is not None for v in [sources, motivation, process]]
    assert all(defined), (
        "Need to define sources, motivation and process in"
        " order to write to this volume."
        "\nDEFINED:"
        f" sources: {defined[0]}"
        f", motivation: {defined[1]}"
        f", process: {defined[2]}"
    )

    # Checking sources
    assert isinstance(sources, list)
    assert all(isinstance(v, str) for v in sources)

    # Checking motivation
    assert isinstance(motivation, str)

    # Checking process
    assert isinstance(process, ptb.Process)


def add_sources(
    cloudvolume: cv.CloudVolume, sources: Optional[list[str]] = None
) -> None:
    """Logs sources in a CloudVolume if they don't already exist.

    Args:
        cloudvolume: A CloudVolume.
        sources: A list of cloudpaths or freeform sources.
    """
    currentsources = set(cloudvolume.provenance.sources)

    newsources = currentsources.union(sources)

    if len(newsources) > len(currentsources):
        cloudvolume.provenance.sources = list(newsources)
        cloudvolume.commit_provenance()


def add_motivation(
    cloudvolume: cv.CloudVolume, motivation: Optional[str] = None
) -> None:
    """Logs a motivation in a CloudVolume if it doesn't already exist.

    Args:
        cloudvolume: A CloudVolume.
        motivation: A reason why this CloudVolume was created.
    """
    if ptb.note_absent(cloudvolume, motivation, ptb.NoteType.MOTIVATION):
        ptb.addmotivation(cloudvolume, motivation)


def add_process(
    cloudvolume: cv.CloudVolume, process: Optional[ptb.Process] = None
) -> None:

    """Logs a ptb.Process in a CloudVolume if not already logged.

    Args:
        cloudvolume: A CloudVolume.
        process: A processing step.
    """
    if ptb.process_absent(cloudvolume, process):
        ptb.logprocess(cloudvolume, process)
    else:
        warnings.warn(
            "Process with the same description already logged." " Skipping logging."
        )


def documentvolume(
    cloudvolume: cv.CloudVolume,
    sources: Optional[list[str]] = None,
    motivation: Optional[str] = None,
    process: Optional[ptb.Process] = None,
) -> None:
    """Logs all required fields for writing to a CirrusVolume.

    Args:
        cloudvolume: A CloudVolume.
        sources: A list of cloudpaths or freeform sources.
        motivation: A reason why this CloudVolume was created.
        process: A processing step.
    """
    add_sources(cloudvolume, sources)
    add_motivation(cloudvolume, motivation)
    add_process(cloudvolume, process)



"""Tests for the high-level CloudVolume wrapper.

Basically just tests whether creating a CloudVolume still works with the
current installation.
"""
import pytest

import cirrusvolume as cv


def test_creation(testcloudvolume):
    """Testing whether basic instantiation of a CirrusVolume works."""
    cirrusvolume = cv.CloudVolume(testcloudvolume.cloudpath)
    assert type(cirrusvolume) == cv.precomputed.CirrusVolumePrecomputed


def test_writeblocking(readvolume):
    """Tests whether writes that don't follow the rules raise an exception."""
    with pytest.raises(AssertionError):
        readvolume[0:10, 0:10, 0:10] = 0



"""A wrapper for CloudVolumePrecomputed."""
from __future__ import annotations

from typing import Optional

import cloudvolume as cv
from cloudvolume.frontends.precomputed import CloudVolumePrecomputed
import provenancetoolbox as ptb

from . import rules
from .volume import register_plugin


def register() -> None:
    """Registration function for CirrusVolumePrecomputed.

    See volume.register_plugin for more information.
    """
    register_plugin(CloudVolumePrecomputed, CirrusVolumePrecomputed)


class CirrusVolumePrecomputed(CloudVolumePrecomputed):
    """A wrapper for CloudVolumePrecomputed.

    See cloudvolume.CloudVolumePrecomputed for inherited methods and
    attributes.

    Additional attributes:
        sources: What volumes/processes were used as inputs to create this
            volume.
        motivation: The reason for creating or modifying this volume.
        process: A provenancetoolbox.Process describing the modifications
            being performed on this volume.
    """

    def __init__(
        self,
        cloudvolume: cv.CloudVolume,
        sources: Optional[list[str]] = None,
        motivation: Optional[str] = None,
        process: Optional[ptb.Process] = None,
    ):
        # Copying the CloudVolume attributes
        self.config = cloudvolume.config
        self.cache = cloudvolume.cache
        self.meta = cloudvolume.meta

        self.image = cloudvolume.image
        self.mesh = cloudvolume.mesh
        self.skeleton = cloudvolume.skeleton

        self.green_threads = cloudvolume.green_threads

        self.mip = cloudvolume.mip
        self.pid = cloudvolume.pid

        # CirrusVolume-specific atrributes
        self.sources = sources
        self.motivation = motivation
        self.process = process

    # Overriding all methods that allow writing to the image
    # with versions that check the rules first
    def __setitem__(self, slices, img):
        self.document()

        super().__setitem__(slices, img)

    def upload_from_shared_memory(
        self, location, bbox, order="F", cutout_bbox=None
    ) -> None:
        self.document()

        super().upload_from_shared_memory(
            location, bbox, order=order, cutout_bbox=cutout_bbox
        )

    def upload_from_file(self, location, bbox, order="F", cutout_bbox=None) -> None:
        self.document()

        super().upload_from_file(location, bbox, order=order, cutout_bbox=cutout_bbox)

    def document(self) -> None:
        """Commits the CirrusVolume documentation to the provenance file."""
        rules.check_writing_rules(self.sources, self.motivation, self.process)

        rules.documentvolume(self, self.sources, self.motivation, self.process)
